#include "include/rcore.h"
#include "include/dataframe.h"
#include "include/covariance.h"
#include "include/blas.h"
#include "include/matrix.h"
#include "include/globals.h"

static void fast_cglm0(SEXP result, double *y, int *z, int n, SEXP confnames,
    int keep) {

int i = 0, nconfigs = length(confnames), *nz = NULL;
double *cc = NULL, *ss = NULL, *ff = NULL, *rr = NULL;
long double *means = NULL, *ssr = NULL;
SEXP coefficients, sd, residuals, fitted;

  /* the matrix of regression coefficients contains the conditional mean given
   * each configuration; estimate in a single pass (set to NaN if there a no
   * observations for a particular observation). */
  PROTECT(coefficients = allocMatrix(REALSXP, 1, nconfigs));
  cc = REAL(coefficients);
  means = Calloc1D(nconfigs, sizeof(long double));
  ssr = Calloc1D(nconfigs, sizeof(long double));
  nz = Calloc1D(nconfigs, sizeof(int));

  for (i = 0; i < n; i++) {
 
    means[z[i] - 1] += y[i];
    nz[z[i] - 1]++;

  }/*FOR*/

  for (i = 0; i < nconfigs; i++) {

    if (nz[i] == 0) {

      means[i] = cc[i] = R_NaN;

    }/*THEN*/
    else {

      means[i] /= nz[i];
      cc[i] = means[i];

    }/*ELSE*/

  }/*FOR*/

  /* compute the fitted values and the residuals (or not), and the standard
   * errors. */
  if (keep) {

    PROTECT(fitted = allocVector(REALSXP, n));
    ff = REAL(fitted);
    for (i  = 0; i < n; i++)
      ff[i] = means[z[i] - 1];

    PROTECT(residuals = allocVector(REALSXP, n));
    rr = REAL(residuals);
    for (i  = 0; i < n; i++)
      rr[i] = y[i] - means[z[i] - 1];

    PROTECT(sd = allocVector(REALSXP, nconfigs));
    ss = REAL(sd);

    for (i = 0; i < n; i++)
      ssr[z[i] - 1] += rr[i] * rr[i];
    for (i = 0; i < nconfigs; i++)
      ss[i] = (nz[i] == 0) ? R_NaN : sqrt(ssr[i] / (nz[i] - 1));

  }/*THEN*/
  else {

    /* fitted values and residuals are just dummy NAs. */
    fitted = residuals = ScalarReal(NA_REAL);

    PROTECT(sd = allocVector(REALSXP, nconfigs));
    ss = REAL(sd);

    for (i = 0; i < n; i++)
      ssr[z[i] - 1] += (y[i] - means[z[i] - 1]) * (y[i] - means[z[i] - 1]);
    for (i = 0; i < nconfigs; i++)
      ss[i] = (nz[i] == 0) ? R_NaN : sqrt(ssr[i] / (nz[i] - 1));

  }/*ELSE*/

  Free1D(means);
  Free1D(ssr);
  Free1D(nz);

  /* save the results and return. */
  setDimNames(coefficients, mkString("(Intercept)"), confnames);
  setAttrib(sd, R_NamesSymbol, confnames);
  SET_VECTOR_ELT(result, 0, coefficients);
  SET_VECTOR_ELT(result, 1, sd);
  SET_VECTOR_ELT(result, 3, residuals);
  SET_VECTOR_ELT(result, 4, fitted);

  UNPROTECT(2 + 2 * keep);

}/*FAST_CGLM0*/

static void fast_cglm1(SEXP result, SEXP parents, double *y, double *x, int *z,
  int n, SEXP confnames, int keep) {

int i = 0, nconfigs = length(confnames), *nz = NULL;
double *cc = NULL, *ff = NULL, *rr = NULL, *ss = NULL, yhat = 0;
long double *mean_y = NULL, *mean_x = NULL, *var_x = NULL, *cov = NULL, *ssr = NULL;
SEXP coefficients, sd, residuals, fitted, coefnames;

  /* the matrix of regression coefficients contains the intercept an the (only)
   * regression coefficient for each configuration; estimate in a two passes. */
  PROTECT(coefficients = allocMatrix(REALSXP, 2, nconfigs));
  cc = REAL(coefficients);

  /* first pass: compute the conditional means of y and x. */
  mean_y = Calloc1D(nconfigs, sizeof(long double));
  mean_x = Calloc1D(nconfigs, sizeof(long double));
  var_x = Calloc1D(nconfigs, sizeof(long double));
  cov = Calloc1D(nconfigs, sizeof(long double));
  ssr = Calloc1D(nconfigs, sizeof(long double));
  nz = Calloc1D(nconfigs, sizeof(int));

  for (i = 0; i < n; i++) {

    mean_y[z[i] - 1] += y[i];
    mean_x[z[i] - 1] += x[i];
    nz[z[i] - 1]++;

  }/*FOR*/

  for (i = 0; i < nconfigs; i++) {

    mean_y[i] /= nz[i];
    mean_x[i] /= nz[i];

  }/*FOR*/


  /* second pass: compute the covariance between x and y and the variance of x. */
  for (i = 0; i < n; i++) {

    var_x[z[i] - 1] += (x[i] - mean_x[z[i] - 1]) * (x[i] - mean_x[z[i] - 1]);
    cov[z[i] - 1] += (y[i] - mean_y[z[i] - 1]) * (x[i] - mean_x[z[i] - 1]);

  }/*FOR*/

  /* the regression coefficients are computed using the closed form estimators
   * for simple regression (set to NaN if there a no observations for a
   * particular observation). */
  for (i = 0; i < nconfigs; i++) {

    if (nz[i] == 0) {

      cc[i * 2] = cc[i * 2 + 1] = R_NaN;

    }/*THEN*/
    else {

      cc[i * 2 + 1] = (fabs(var_x[i]) < MACHINE_TOL) ? 0 : cov[i] / var_x[i];
      cc[i * 2] = mean_y[i] - mean_x[i] * cc[i * 2 + 1];

    }/*ELSE*/

  }/*FOR*/

  /* compute the fitted values and the residuals (or not), and the standard
   * errors. */
  if (keep) {

    PROTECT(fitted = allocVector(REALSXP, n));
    ff = REAL(fitted);
    for (i  = 0; i < n; i++)
      ff[i] = cc[(z[i] - 1) * 2] + x[i] * cc[(z[i] - 1) * 2 + 1];

    PROTECT(residuals = allocVector(REALSXP, n));
    rr = REAL(residuals);
    for (i  = 0; i < n; i++)
      rr[i] = y[i] - ff[i];

    PROTECT(sd = allocVector(REALSXP, nconfigs));
    ss = REAL(sd);

    for (i = 0; i < n; i++)
      ssr[z[i] - 1] += rr[i] * rr[i];
    for (i = 0; i < nconfigs; i++)
      ss[i] = (nz[i] == 0) ? R_NaN : sqrt(ssr[i] / (nz[i] - 2));

  }/*THEN*/
  else {

    /* fitted values and residuals are just dummy NAs. */
    fitted = residuals = ScalarReal(NA_REAL);

    PROTECT(sd = allocVector(REALSXP, nconfigs));
    ss = REAL(sd);

    for (i = 0; i < n; i++) {

      yhat = cc[(z[i] - 1) * 2] + x[i] * cc[(z[i] - 1) * 2 + 1];
      ssr[z[i] - 1] += (y[i] - yhat) * (y[i] - yhat);

    }/*FOR*/
    for (i = 0; i < nconfigs; i++)
      ss[i] = (nz[i] == 0) ? R_NaN : sqrt(ssr[i] / (nz[i] - 2));

  }/*ELSE*/

  Free1D(mean_y);
  Free1D(mean_x);
  Free1D(var_x);
  Free1D(cov);
  Free1D(ssr);
  Free1D(nz);

  /* save the results and return. */
  PROTECT(coefnames = allocVector(STRSXP, 2));
  SET_STRING_ELT(coefnames, 0, mkChar("(Intercept)"));
  SET_STRING_ELT(coefnames, 1, STRING_ELT(parents, 0));
  setDimNames(coefficients, coefnames, confnames);
  setAttrib(sd, R_NamesSymbol, confnames);
  SET_VECTOR_ELT(result, 0, coefficients);
  SET_VECTOR_ELT(result, 1, sd);
  SET_VECTOR_ELT(result, 3, residuals);
  SET_VECTOR_ELT(result, 4, fitted);

  UNPROTECT(3 + 2 * keep);

}/*FAST_CGLM1*/

static void fast_cglmp(SEXP result, SEXP data, SEXP parents, double *y,
  int *z, int n, SEXP confnames, int keep) {

int i = 0, j = 0, *counter = NULL;
int nconfigs = length(confnames), *nz = NULL, ncol = length(parents) + 1;
double **qr_matrix = NULL, **column = NULL, *ffz = NULL, *rrz = NULL;
double **yy = NULL, *cc = NULL, *ss = NULL, *ff = NULL, *rr = NULL;
SEXP coefficients, sd, residuals, fitted, coefnames, regressors;

  /* first pass: count how many observations for each configuration. */
  nz = Calloc1D(nconfigs, sizeof(int));
  for (int i = 0; i < n; i++)
    nz[z[i] - 1]++;

  /* allocate the responses and the matrices to pass to c_qr_ols(). */
  yy = Calloc1D(nconfigs, sizeof(double *));
  for (i = 0; i < nconfigs; i++)
    yy[i] = Calloc1D(nz[i], sizeof(double));

  qr_matrix = Calloc1D(nconfigs, sizeof(double *));
  for (i = 0; i < nconfigs; i++)
    qr_matrix[i] = Calloc1D(nz[i] * ncol, sizeof(double));

  /* set the intercept. */
  for (i = 0; i < nconfigs; i++)
   for (j = 0; j < nz[i]; j++)
     qr_matrix[i][j] = 1;

  /* dereference the relevant variables in the data. */
  PROTECT(regressors = c_dataframe_column(data, parents, FALSE, FALSE));
  column = Calloc1D(ncol - 1, sizeof(double *));
  for (i = 0; i < ncol - 1; i++)
    column[i] = REAL(VECTOR_ELT(regressors, i));

  /* second pass: copy the data into the matrices. */
  counter = Calloc1D(nconfigs, sizeof(int));

  for (i = 0; i < n; i++) {

    for (j = 1; j < ncol; j++)
      qr_matrix[z[i] - 1][CMC(counter[z[i] - 1], j, nz[z[i] - 1])] = column[j - 1][i];

    yy[z[i] - 1][counter[z[i] - 1]] = y[i];

    counter[z[i] - 1]++;

  }/*FOR*/

  /* allocate the regression coefficients and the standard errors. */
  PROTECT(coefficients = allocMatrix(REALSXP, ncol, nconfigs));
  cc = REAL(coefficients);
  PROTECT(sd = allocVector(REALSXP, nconfigs));
  ss = REAL(sd);

  /* compute the fitted values and the residuals (or not). */
  if (keep) {

    memset(counter, '\0', nconfigs * sizeof(int));
    PROTECT(fitted = allocVector(REALSXP, n));
    ff = REAL(fitted);
    PROTECT(residuals = allocVector(REALSXP, n));
    rr = REAL(residuals);

    for (i = 0; i < nconfigs; i++) {

      if (nz[i] == 0) {

        for (j = 0; j < ncol; j++)
          cc[i * ncol + j] = R_NaN;
        ss[i] = R_NaN;
        continue;

      }/*THEN*/

      ffz = Calloc1D(nz[i], sizeof(double));
      rrz = Calloc1D(nz[i], sizeof(double));

      /* estimate OLS using the QR decomposition. */
      c_qr_ols(qr_matrix[i], yy[i], nz[i], ncol, ffz, rrz, 
        cc + i * ncol, ss + i);

      for (j = 0; j < n; j++) {

        if (z[j] - 1 == i) {

          ff[j] = ffz[counter[i]];
          rr[j] = rrz[counter[i]++];

        }/*THEN*/

      }/*FOR*/

      Free1D(ffz);
      Free1D(rrz);

    }/*FOR*/

  }/*THEN*/
  else {

    /* fitted values and residuals are just dummy NAs. */
    fitted = residuals = ScalarReal(NA_REAL);

    for (i = 0; i < nconfigs; i++) {

      /* estimate OLS using the QR decomposition. */
      c_qr_ols(qr_matrix[i], yy[i], nz[i], ncol, NULL, NULL, 
        cc + i * ncol, ss + i);

    }/*FOR*/

  }/*ELSE*/

  Free2D(yy, nconfigs);
  Free1D(nz);
  Free1D(counter);
  Free1D(column);
  Free2D(qr_matrix, nconfigs);

  /* save the results and return. */
  PROTECT(coefnames = allocVector(STRSXP, ncol));
  SET_STRING_ELT(coefnames, 0, mkChar("(Intercept)"));
  for (i = 1; i < ncol; i++)
    SET_STRING_ELT(coefnames, i, STRING_ELT(parents, i - 1));
  setDimNames(coefficients, coefnames, confnames);
  setAttrib(sd, R_NamesSymbol, confnames);
  SET_VECTOR_ELT(result, 0, coefficients);
  SET_VECTOR_ELT(result, 1, sd);
  SET_VECTOR_ELT(result, 3, residuals);
  SET_VECTOR_ELT(result, 4, fitted);

  UNPROTECT(4 + 2 * keep);

}/*FAST_CLGMP*/

SEXP fast_cglm(SEXP data, SEXP node, SEXP parents, SEXP configs, SEXP keep) {

int n = 0, *z = NULL;
double *y = NULL;
SEXP result, response, confnames, dummy_configs;

  /* dereference the response variable and the configurations. */
  PROTECT(response = c_dataframe_column(data, node, TRUE, FALSE));
  y = REAL(response);
  n = length(response);
  z = INTEGER(configs);
  /* prepare the configuration index names. */
  confnames = getAttrib(configs, R_LevelsSymbol);
  /* allocate and initialize the return value. */
  PROTECT(result = allocVector(VECSXP, 5));
  setAttrib(result, R_NamesSymbol,
    mkStringVec(5, "coefficients", "sd", "configs", "residuals", "fitted.values"));
  if (isTRUE(keep)) {

    SET_VECTOR_ELT(result, 2, configs);

  }/*THEN*/
  else {

    PROTECT(dummy_configs = allocVector(INTSXP, 1));
    INT(dummy_configs) = NA_INTEGER;
    setAttrib(dummy_configs, R_ClassSymbol, mkString("factor"));
    setAttrib(dummy_configs, R_LevelsSymbol, confnames);
    SET_VECTOR_ELT(result, 2, dummy_configs);
    UNPROTECT(1);

  }/*ELSE*/
 
  if (length(parents) == 0) {

    /* the regression includes only the intercept of the model, which the mean
     * of the response. */
    fast_cglm0(result, y, z, n, confnames, isTRUE(keep));

  }/*THEN*/
  else if (length(parents) == 1) {

    /* the regression includes only the intercept and one regression coefficient,
     * so the solution is in closed form (simple linear regression case). */
    double *x = REAL(c_dataframe_column(data, parents, TRUE, FALSE));

    fast_cglm1(result, parents, y, x, z, n, confnames, isTRUE(keep));

  }/*THEN*/
  else {

    /* general multiple regression, computed using QR decomposition. */
    fast_cglmp(result, data, parents, y, z, n, confnames, isTRUE(keep));

  }/*ELSE*/

  UNPROTECT(2);

  return result;

}/*FAST_LM*/

