#include "include/rcore.h"
#include "include/dataframe.h"
#include "include/covariance.h"
#include "include/blas.h"
#include "include/globals.h"

/* fast implementation of a linear regression with just the intercept. */
static void fast_lm0(SEXP result, double *y, int n, int keep) {

int i = 0;
double mean = c_mean(y, n);
double sse = c_sse(y, mean, n);
double *ff = NULL, *rr = NULL;
SEXP coefficients, sd, residuals, fitted;

  /* the only coefficient is the intercept, which is the mean of the response. */
  PROTECT(coefficients = ScalarReal(mean));
  setAttrib(coefficients, R_NamesSymbol, mkString("(Intercept)"));
  /* again the standerd error is just the standard deviation of the response. */
  PROTECT(sd = ScalarReal(sqrt(sse / (n - 1))));

  /* compute the fitted values and the residuals (or not). */
  if (keep) {

    PROTECT(fitted = allocVector(REALSXP, n));
    ff = REAL(fitted);
    for (i  = 0; i < n; i++)
      ff[i] = mean;

    PROTECT(residuals = allocVector(REALSXP, n));
    rr = REAL(residuals);
    for (i  = 0; i < n; i++)
      rr[i] = y[i] - mean;

  }/*THEN*/
  else {

    /* fitted values and residuals are just dummy NAs. */
    fitted = residuals = ScalarReal(NA_REAL);

  }/*ELSE*/

  /* save the results and return. */
  SET_VECTOR_ELT(result, 0, coefficients);
  SET_VECTOR_ELT(result, 1, sd);
  SET_VECTOR_ELT(result, 2, residuals);
  SET_VECTOR_ELT(result, 3, fitted);

  UNPROTECT(2 + 2 * keep);

}/*FAST_LM0*/

/* fast implementation of a simple linear regression. */
static void fast_lm1(SEXP result, SEXP parents, double *y, double *x, int n,
    int keep) {

int i = 0;
double mean[2] = {0, 0}, cov[4] = {0, 0, 0, 0}, mu = 0, beta = 0, sse = 0;
double *m[2] = {y, x};
double *ff = NULL, *rr = NULL;
SEXP coefficients, sd, residuals, fitted, coefnames;

  /* the regression coefficients are computed using the closed form estimators
   * for simple regression. */
  c_meanvec(m, mean, n, 2, 0);
  c_covmat(m, mean, 2, n, cov, 0);
  beta = (fabs(cov[3]) < MACHINE_TOL) ? 0 : cov[1] / cov[3];
  mu = mean[0] - mean[1] * beta;

  PROTECT(coefficients = allocVector(REALSXP, 2));
  REAL(coefficients)[0] = mu;
  REAL(coefficients)[1] = beta;
  PROTECT(coefnames = allocVector(STRSXP, 2));
  SET_STRING_ELT(coefnames, 0, mkChar("(Intercept)"));
  SET_STRING_ELT(coefnames, 1, STRING_ELT(parents, 0));
  setAttrib(coefficients, R_NamesSymbol, coefnames);

  /* compute the fitted values and the residuals (or not). */
  if (keep) {

    PROTECT(fitted = allocVector(REALSXP, n));
    ff = REAL(fitted);
    for (i = 0; i < n; i++)
      ff[i] = mu + m[1][i] * beta;

    PROTECT(residuals = allocVector(REALSXP, n));
    rr = REAL(residuals);
    for (i = 0; i < n; i++)
      rr[i] = y[i] - ff[i];

    /* compute the standard error from the residuals. */
    for (i = 0; i < n; i++)
      sse += rr[i] * rr[i];
    PROTECT(sd = ScalarReal(sqrt(sse / (n - 2))));

  }/*THEN*/
  else {

    /* fitted values and residuals are just dummy NAs. */
    fitted = residuals = ScalarReal(NA_REAL);
    /* compute the standard error from the residuals. */
    for (i = 0; i < n; i++)
      sse += (y[i] - (mu + m[1][i] * beta)) * (y[i] - (mu + m[1][i] * beta));
    PROTECT(sd = ScalarReal(sqrt(sse / (n - 2))));

  }/*ELSE*/

  /* save the results and return. */
  SET_VECTOR_ELT(result, 0, coefficients);
  SET_VECTOR_ELT(result, 1, sd);
  SET_VECTOR_ELT(result, 2, residuals);
  SET_VECTOR_ELT(result, 3, fitted);

  UNPROTECT(3 + 2 * keep);

}/*FAST_LM1*/

static void fast_lmp(SEXP result, SEXP data, SEXP parents, double *y, int n,
    int keep) {

int i = 0, ncol = length(parents) + 1;
double *qr = NULL;
SEXP qr_x, coefficients, sd, residuals, fitted, coefnames;

  /* prepare the data for the QR decomposition. */
  PROTECT(qr_x = qr_matrix(data, parents));
  qr = REAL(qr_x);
  /* allocate the coefficients and the standard error. */
  PROTECT(coefficients = allocVector(REALSXP, ncol));
  PROTECT(coefnames = allocVector(STRSXP, ncol));
  SET_STRING_ELT(coefnames, 0, mkChar("(Intercept)"));
  for (i = 1; i < ncol; i++)
    SET_STRING_ELT(coefnames, i, STRING_ELT(parents, i - 1));
  setAttrib(coefficients, R_NamesSymbol, coefnames);
  PROTECT(sd = allocVector(REALSXP, 1));

  /* compute the fitted values and the residuals (or not). */
  if (keep) {

    /* fitted values and residuals are extracted and returned. */
    PROTECT(fitted = allocVector(REALSXP, n));
    PROTECT(residuals = allocVector(REALSXP, n));
    /* estimate OLS using the QR decomposition. */
    c_qr_ols(qr, y, n, ncol, REAL(fitted), REAL(residuals),
      REAL(coefficients), REAL(sd));

  }/*THEN*/
  else {

    /* fitted values and residuals are just dummy NAs. */
    fitted = residuals = ScalarReal(NA_REAL);
    /* estimate OLS using the QR decomposition. */
    c_qr_ols(qr, y, n, ncol, NULL, NULL, REAL(coefficients), REAL(sd));

  }/*ELSE*/

  /* save the results and return. */
  SET_VECTOR_ELT(result, 0, coefficients);
  SET_VECTOR_ELT(result, 1, sd);
  SET_VECTOR_ELT(result, 2, residuals);
  SET_VECTOR_ELT(result, 3, fitted);

  UNPROTECT(4 + 2 * keep);

}/*FAST_LMP*/

SEXP fast_lm(SEXP data, SEXP node, SEXP parents, SEXP keep) {

int n = 0;
double *y = NULL;
SEXP result, response;

  /* dereference the response variable. */
  PROTECT(response = c_dataframe_column(data, node, TRUE, FALSE));
  y = REAL(response);
  n = length(response);
  /* allocate and initialize the return value. */
  PROTECT(result = allocVector(VECSXP, 4));
  setAttrib(result, R_NamesSymbol,
    mkStringVec(4, "coefficients", "sd", "residuals", "fitted.values"));

  if (length(parents) == 0) {

    /* the regression includes only the intercept of the model, which the mean
     * of the response. */
    fast_lm0(result, y, n, isTRUE(keep));

  }/*THEN*/
  else if (length(parents) == 1) {

    /* the regression includes only the intercept and one regression coefficient,
     * so the solution is in closed form (simple linear regression case). */
    double *x = REAL(c_dataframe_column(data, parents, TRUE, FALSE));

    fast_lm1(result, parents, y, x, n, isTRUE(keep));

  }/*THEN*/
  else {

    /* general multiple regression, computed using QR decomposition. */
    fast_lmp(result, data, parents, y, n, isTRUE(keep));

  }/*ELSE*/

  UNPROTECT(2);

  return result;

}/*FAST_LM*/
